/****************************************************
  Simple Compiler
****************************************************/
#include <stdlib.h>
#include <stdio.h>
#include <stdbool.h>
#include <string.h>
#include <ctype.h>
#include <stdint.h>
#include <unistd.h>

// Includes constants, CPU model and some helper functions
#include "SimpleC.h"

/* SIMPLE OpCode INTERPRETER */
void Simple(void); // Forward reference
#include "SimpleInt.c"

/* SIMPLE OpCode LISTER */
void List(void); // Forward reference
#include "SimpleOpCode.c"

/* SYMBOL TABLE */
const int MaxSymbol=1000;
char symbolName[MaxSymbol][16]; // Keywords and Identifiers have a Name
char symbolTok[MaxSymbol];      // Everything has a Token constant
int symbolAdr[MaxSymbol];       // If an Identifier then it has an Address
int nSymbol=0;

void AddIdentifier(char *Name)
{
  strcpy(symbolName[nSymbol],Name);
  symbolTok[nSymbol]=_ident;
  symbolAdr[nSymbol]=DP;
  DP=DP+1;
  nSymbol++;
}

int FindName(char *Name)
{
  int p;

  for (p=1;p<=nSymbol;p++) {
    if (strcmp(symbolName[p-1],Name)==0) break;
  }
  if (p>nSymbol) {
    return 0; // false
  } else {
    return p; // true - return index (1 .. nSymbol)
  }
}

int GetKey(char *Name)
{
  int p;

  p=FindName(Name);
  if (p>0) return symbolTok[p-1];
  return 0;
}

bool FindKey(int key, char *Name)
{
  int p;

  for (p=1;p<=nSymbol;p++) {
    if (symbolTok[p-1]==key) break;
  }
  if (p>nSymbol) {
    strcpy(Name,"");
    return false;
  } else {
    strcpy(Name,symbolName[p-1]);
    return true;
  }
}

bool FindVar(char *Name, int *Adr)
{
  int p;

  p=FindName(Name);
  if (p>0) {
    *Adr=symbolAdr[p-1];
    return true;  // Name found (includes keywords)
  } else {
    return false; // Name not found
  }
}

/* TOKENISER */
bool Error=false;
char ErrorMsg[60]="";
char ch;
char tTok;
char sTok[16]="";
int iTok;

char nextchar(void)
{
  return toupper(getchar());
}

void GetIdent(void)
{
  int p;

  while (isalpha(ch)) {
    // Append ch to sTok
    p=strlen(sTok);
    sTok[p]=ch;
    sTok[p+1]='\0';
    ch=nextchar();
  }
  tTok=GetKey(sTok);        // Existing keyword or identifier
  if (tTok==0) tTok=_ident; // New identifier
}

void GetNum(void)
{
  int p;

  while (isdigit(ch)) {
    // Append ch to sTok
    p=strlen(sTok);
    sTok[p]=ch;
    sTok[p+1]='\0';
    ch=nextchar();
  }
  iTok=atoi(sTok);
  tTok=_integer;
}

void GetNextToken(void)
{
  static bool newline=false;
  static int nline=0;

  // Print to sterr the tokenized data
  if (nline==0) {
    nline++;
    fprintf(stderr,"Tokenised Code:\nLine %6d:",nline);
  } else if (newline) {
    if (Error) {
      // Present error then exit
      fprintf(stderr,"\nError on last line: %s\n",ErrorMsg);
      exit(1);
    }
    newline=false;
    nline++;
    fprintf(stderr,"\nLine %6d: %s",nline,sTok);
  } else {
    fprintf(stderr," %s",sTok);
  }

  // Reset tokens
  strcpy(sTok,"");
  iTok=0;
  tTok=_unknown;

  // Skip white spaces
  while (isspace(ch)||iscntrl(ch)) {
    // Test for new line
    if (ch=='\n') newline=true;
    ch=nextchar();
  }
  if (isalpha(ch)) {
    GetIdent();
  } else if (isdigit(ch)) {
    GetNum();
  } else if (ch==',') {
    tTok=_comma;
    ch=nextchar();
  } else if (ch=='+') {
    tTok=_plus;
    ch=nextchar();
  } else if (ch=='-') {
    tTok=_minus;
    ch=nextchar();
  } else if (ch=='*') {
    tTok=_mul;
    ch=nextchar();
  } else if (ch=='/') {
    tTok=_div;
    ch=nextchar();
  } else if (ch=='(') {
    tTok=_lparent;
    ch=nextchar();
  } else if (ch==')') {
    tTok=_rparent;
    ch=nextchar();
  } else if (ch=='=') {
    tTok=_eq;
    ch=nextchar();
  } else if (ch=='<') {
    // Includes <, <= and <>
    tTok=_lt;
    ch=nextchar();
    if (ch=='=') {
      tTok=_le;
      ch=nextchar();
    } else if (ch== '>') {
      tTok=_ne;
      ch=nextchar();
    }
  } else if (ch=='>') {
    // Includes > and >=
    tTok=_gt;
    ch=nextchar();
    if (ch=='=') {
      tTok=_ge;
      ch=nextchar();
    }
  } else if (ch==EOF) {
    // End of File - Exit gracefully
    tTok=_halt;
    strcpy(sTok,"halt");
    fprintf(stderr,"\n");
  } else {
    tTok=_unknown;
    ch=nextchar();
  }
  // Set all token values
  if (tTok==_integer) {
    itoa(iTok,sTok,10);
  } else if (tTok>_integer) {
    if (!FindName(sTok)) FindKey(tTok,sTok);
  }
}

/* COMPILER */
int IP=0;           // Instruction pointer
void Emit(int data)
{
  Memory[IP]=data;
  IP=IP+1;
}

void Emit2(int code, int addr)
{
  Emit(code);
  Emit(addr);
}

void FixUp(int adr, int addr)
{
  Memory[adr+1]=addr;
}

void Halt(void)
{
  Emit(_halt);
}

void BoolExpression(void); // Forward reference

void Factor(void)
{
  int adr=0;

  if (tTok==_ident) {
    if (!FindVar(sTok, &adr)) {
      AddIdentifier(sTok);
      FindVar(sTok, &adr);
    }
    Emit2(_movAxVar, adr);
  } else if (tTok==_integer) {
    // Load const
    Emit2(_movAxImm, iTok);
  } else if (tTok==_lparent) {
    GetNextToken();
    BoolExpression();
    if (tTok!=_rparent) {
      // Error
      Error=true;
      strcpy(ErrorMsg,"Expected ')'");
    }
  } else {
    // Error
    Error=true;
    strcpy(ErrorMsg,"Unknown Expression");
  }
  GetNextToken();
}

void Term(void)
{
  char op;

  Factor();
  while ((tTok==_mul)||(tTok==_div)) {
    op=tTok;
    Emit(_pushAx);
    GetNextToken();
    Factor();
    if (op==_mul) {
      Emit(_popBx);
      Emit(_mulBx);
    } else if (op==_div) {
      Emit(_movBxAx);
      Emit(_popAx);
      Emit(_divBx);
    }
  }
}

void Expression(void)
{
  char op;

  if ((tTok==_plus)||(tTok==_minus)) {
    Emit(_xorAxAx);
  } else {
    Term();
  }
  while ((tTok==_plus)||(tTok==_minus)) {
    op=tTok;
    Emit(_pushAx);
    GetNextToken();
    Term();
    if (op==_plus ){
      Emit(_popBx);
      Emit(_addBx);
    } else if(op==_minus) {
      Emit(_movBxAx);
      Emit(_popAx);
      Emit(_subBx);
    }
  }
}

void BoolExpression(void)
{
  char op;

  Expression();
  if ((tTok==_eq)||(tTok==_ne)||(tTok==_lt)||(tTok==_le)||(tTok==_gt)||(tTok==_ge)) {
    op=tTok;
    Emit(_pushAx);
    GetNextToken();
    Expression();
    Emit(_movBxAx);
    Emit(_popAx);
    Emit(_cmpBx);
    if (op==_eq) {
      Emit(_setEq);
    } else if (op==_ne) {
      Emit(_setNE);
    } else if (op==_lt) {
      Emit(_setLT);
    } else if (op==_le) {
      Emit(_setLE);
    } else if (op==_gt) {
      Emit(_setGT);
    } else if (op==_ge) {
      Emit(_setGE);
    }
  }
}

void Print(void)
{
  GetNextToken();
  BoolExpression();
  Emit(_wrtAx);
  while (tTok==_comma) {
    GetNextToken();
    BoolExpression();
    Emit(_wrtAx);
  }
  Emit(_wrtLn);
}

void Input(void)
{
  Emit(_readAx);
}

void Assignment(void)
{
  int adr=0;

  if (!FindVar(sTok, &adr)) {
    AddIdentifier(sTok);
    FindVar(sTok, &adr);
  }
  GetNextToken();
  if (tTok==_eq) {
    GetNextToken();
  } else {
    // Error
    Error=true;
    strcpy(ErrorMsg,"Expected '='");
  }
  BoolExpression();
  Emit2(_movVarAx, adr);
}

void Statement(void); // Forward reference

void WhileState(void)
{
  int pc1,pc2;

  GetNextToken();
  pc1=IP;
  BoolExpression();
  Emit(_orAxAx);
  pc2=IP;
  Emit2(_jz, 0);
  Statement();
  Emit2(_jmp,pc1);
  FixUp(pc2, IP);
}

void IfState(void)
{
  int pc1,pc2;

  GetNextToken();
  BoolExpression();
  Emit(_orAxAx);
  pc1=IP;
  Emit2(_jz, 0);
  if (tTok==_then) {
    GetNextToken();
  } else {
    // Error
    Error=true;
    strcpy(ErrorMsg,"Expected THEN");
  }
  Statement();
  if (tTok!=_else) {
    FixUp(pc1, IP);
  } else {
    GetNextToken();
    pc2=IP;
    Emit2(_jmp, 0);
    FixUp(pc1, IP);
    Statement();
    FixUp(pc2, IP);
  }
}

void BeginState(void)
{
  GetNextToken();
  while ((tTok!=_end)&&(tTok!=_halt)) Statement();
  GetNextToken(); // Tokeniser will send an _halt if EOF
}

void Statement(void)
{
  if (tTok==_begin) {
    BeginState();
  } else if (tTok==_while) {
    WhileState();
  } else if (tTok==_if) {
    IfState();
  } else if (tTok==_writeAx) {
    Print();
  } else if (tTok==_readAx) {
    Input();
  } else if (tTok==_ident) {
    Assignment();
  } else {
    // Error
    Error=true;
    strcpy(ErrorMsg,"Unknown Statement");
  }
}

void AddKeyword(char *Name, int Id)
{
  strcpy(symbolName[nSymbol],Name);
  symbolTok[nSymbol]=(char)Id; // tTok Id
  symbolAdr[nSymbol]=0;
  nSymbol++;
}

void Init(void)
{
  // Initialise symbol table with keywords
  nSymbol=0;
  strcpy(sTok, "BEGIN");AddKeyword(sTok, _begin);
  strcpy(sTok, "ELSE"); AddKeyword(sTok, _else);
  strcpy(sTok, "END");  AddKeyword(sTok, _end);
  strcpy(sTok, "IF");   AddKeyword(sTok, _if);
  strcpy(sTok, "THEN"); AddKeyword(sTok, _then);
  strcpy(sTok, "WHILE");AddKeyword(sTok, _while);
  strcpy(sTok, "WRITE");AddKeyword(sTok, _writeAx);
  strcpy(sTok, "READ"); AddKeyword(sTok, _readAx);
  // Initialise symbol table with symbols
  strcpy(sTok, "("); AddKeyword(sTok, _lparent);
  strcpy(sTok, ")"); AddKeyword(sTok, _rparent);
  strcpy(sTok, "+"); AddKeyword(sTok, _plus);
  strcpy(sTok, "-"); AddKeyword(sTok, _minus);
  strcpy(sTok, "*"); AddKeyword(sTok, _mul);
  strcpy(sTok, "/"); AddKeyword(sTok, _div);
  strcpy(sTok, ","); AddKeyword(sTok, _comma);
  strcpy(sTok, "="); AddKeyword(sTok, _eq);
  strcpy(sTok, "<>"); AddKeyword(sTok, _ne);
  strcpy(sTok, "<"); AddKeyword(sTok, _lt);
  strcpy(sTok, "<="); AddKeyword(sTok, _le);
  strcpy(sTok, ">"); AddKeyword(sTok, _gt);
  strcpy(sTok, ">="); AddKeyword(sTok, _ge);
  strcpy(sTok, "?"); AddKeyword(sTok, _unknown);

  // Initialise GetNextToken()
  ch=nextchar();
}

void Compile(void)
{
  GetNextToken();
  Statement();
  Halt();

}

int main(int argc, char **argv)
{
  // Option flags
  bool fileOutFlag=false;
  char fileOutName[80]="";
  bool SimpleIntFlag=false;

  // Get parameters
  {
    int k;

    for (k=1;k<argc;k++) {
      if (strcmp(argv[k],"-o")==0) {
        fileOutFlag=true;
        if (k+1<argc) strcpy(fileOutName,argv[k+1]);
      }
      if (strcmp(argv[k],"-int")==0) SimpleIntFlag=true;
    }
  }
  if (fileOutFlag==false) strcpy(fileOutName,"stdout");

  Init();
  Compile();
  SimpleOpCode(fileOutName);
  if (SimpleIntFlag) SimpleInt();
  printf("Done.\n");

  return 0;
}
